"""Index creating step for the alignment."""
import logging
import os
import subprocess
import warnings

import gtfparse
import luigi

from . import util
from .config import GeneralConfig


class BuildBlastDatabase(luigi.Task):
    """Build local nucleotide blast database."""

    logger = logging.getLogger("custom-logger")

    def output(self):
        return [
            luigi.LocalTarget(f"{util.get_genome_name()}.{extension}")
            for extension in ["nhr", "nin", "nsq"]
        ]

    def build_blast_index(self, fname: str, genome: str) -> None:
        args_blast = ["makeblastdb", "-dbtype", "nucl", "-in", fname, "-out", genome]
        self.logger.debug(f"Running bowtie with - {''.join(args_blast)}")
        subprocess.check_call(
            args_blast, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )

    def run(self):
        self.build_blast_index(
            fname=os.path.abspath(GeneralConfig().reference_genome),
            genome=util.get_genome_name(),
        )
        self.logger.info("Finished building blast database.")


class BuildBowtieIndex(luigi.Task):
    """Create bowtie index for a given reference genome."""

    logger = logging.getLogger("custom-logger")

    def output(self):
        return [
            luigi.LocalTarget(util.get_genome_name() + ext)
            for ext in [
                ".1.ebwt",
                ".2.ebwt",
                ".3.ebwt",
                ".4.ebwt",
                ".rev.1.ebwt",
                ".rev.2.ebwt",
            ]
        ]

    def build_bowtie_index(self, fname: str, genome: str) -> None:
        """Build bowtie index for file fname titled genome."""
        args_bowtie = ["bowtie-build", fname, genome]
        self.logger.debug(f"Running bowtie with - {''.join(args_bowtie)}")
        subprocess.check_call(
            args_bowtie, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )

    def run(self):
        self.build_bowtie_index(
            fname=os.path.abspath(GeneralConfig().reference_genome),
            genome=util.get_genome_name(),
        )
        self.logger.info("Finished building bowtie index.")


class PrepareAnnotationFile(luigi.Task):
    """Convert gtf annotation file to parquet.

    Doesn't change any data but greatly speeds up subsequent I/O (~10x increase).
    """

    logger = logging.getLogger("custom-logger")

    def output(self):
        return luigi.LocalTarget(GeneralConfig().reference_annotation + ".parq")

    def prepare_gtf_file(self, fname_input: str, fname_output: str) -> None:
        """Save GTF file as parquet."""
        warnings.filterwarnings("ignore")
        df = gtfparse.read_gtf(fname_input)
        warnings.filterwarnings("default")
        self.logger.debug("Read gtf file with gtfparse")
        df.to_parquet(fname_output)

    def run(self):
        self.prepare_gtf_file(GeneralConfig().reference_annotation, self.output().path)
        self.logger.info("Finished parsing GTF annotation file.")




"""Analyze a probe set and save plots."""

from typing import List, Union
import logging
import multiprocessing
import os

from mpl_toolkits.axes_grid1 import make_axes_locatable
import Bio.Seq
import Bio.SeqIO
import luigi
import matplotlib.pyplot as plt
import numpy as np
import pysam

from . import util
from .alignment import AlignProbeCandidates
from .basic_filtering import get_g_quadruplet_count
from .basic_filtering import get_gc_content
from .basic_filtering import get_melting_temp
from .config import GeneralConfig
from .config import ProbeConfig
from .config import RunConfig
from .config import SequenceConfig
from .indexing import BuildBowtieIndex
from .kmers import BuildJellyfishIndex
from .kmers import get_max_kmer_count
from .prepare_sequence import PrepareSequence
from .secondary_structure import get_free_energy


class AnalyzeProbeset(luigi.Task):
    """Analysis of a probe set with summary stats."""

    logger = logging.getLogger("custom-logger")

    def requires(self):
        tasks = {
            "jellyfish": BuildJellyfishIndex(),
            "bowtie": BuildBowtieIndex(),
            "sequence": PrepareSequence(),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(
            os.path.splitext(os.path.basename(RunConfig().analyze_probeset))[0]
            + "_analysis.pdf"
        )

    @property
    def gene(self):
        """Parse gene sequence file depending on parameters passed."""
        return Bio.SeqIO.read(
            self.input()["sequence"].path, format="fasta"
        ).reverse_complement()

    @staticmethod
    def histplot(ax: plt.Axes, data: List[int], title: str, min_value: int):
        """Basic narrow count based histogram."""
        ax.set_title(title)
        bins = np.arange(min_value, max(data) + 4) + 0.5
        ax.hist(data, bins=bins)
        ax.set(xticks=bins[1:] - 0.5, ylabel="Count")

    @staticmethod
    def boxplot(ax: plt.Axes, data: List[Union[int, float]], title: str, ylabel: str):
        """Boxplot with kde-esque lines next to it."""
        ax.set_title(title)
        ax.boxplot(data)
        ax.eventplot(
            data,
            orientation="vertical",
            linewidths=0.2,
            lineoffsets=0.8,
            linelengths=0.1,
        )
        ax.set(xticks=[], ylabel=ylabel)

    @staticmethod
    def matrix(ax: plt.Axes, data: np.ndarray, title: str, fig):  # type: ignore
        """Image with colorbar in same axis."""
        ax.set_title(title)
        image = ax.imshow(data, vmin=0, vmax=1, cmap="coolwarm")
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(image, cax=cax, orientation="vertical")

    def add_length(self, ax: plt.Axes):
        """Add probe lengths to axis."""
        lengths = list(map(len, self.sequences))
        self.histplot(ax, lengths, "Lengths", min(lengths) - 4)
        self.logger.debug("Added length to axis.")

    def add_melting_temperature(self, ax: plt.Axes):
        """Add probe melting temperatures to axis."""
        melting_temps = [
            get_melting_temp(
                seq.seq,
                ProbeConfig().na_concentration,
                ProbeConfig().formamide_concentration,
            )
            for seq in self.sequences
        ]
        self.boxplot(ax, melting_temps, "Melting temperatures", "ºC")
        self.logger.debug("Added melting temperature to axis.")

    def add_gc_content(self, ax: plt.Axes):
        """Add probe GC content to axis."""
        gc_content = [get_gc_content(seq.seq) for seq in self.sequences]
        self.boxplot(ax, gc_content, "GC Content", "%")
        self.logger.debug("Added GC content to axis.")

    def add_g_quadruplet(self, ax: plt.Axes):
        """Add number of G quadruplets in probes to axis."""
        quads = [get_g_quadruplet_count(seq.seq) for seq in self.sequences]
        self.histplot(ax, quads, "G quadruplet", -1)
        self.logger.debug("Added G quadruplets to axis.")

    def _get_max_kmer_count(self, seq):
        return get_max_kmer_count(seq, self.jellyfish_path)
        # return get_max_kmer_count(seq, "./tests/data/sacCer3_15.jf")

    def add_kmers(self, ax: plt.Axes):
        """Add highest number of sub-kmers in probes to axis."""
        self.jellyfish_path = self.input()["jellyfish"].path
        with multiprocessing.Pool(GeneralConfig().threads) as pool:
            kmers = pool.map(self._get_max_kmer_count, self.sequences)
        self.histplot(ax, kmers, "K-mer count (len)", -1)
        self.logger.debug("Added kmers to axis.")

    def add_free_energy(self, ax: plt.Axes):
        """Add predicted deltaG of probes to axis."""
        with multiprocessing.Pool(GeneralConfig().threads) as pool:
            deltag = pool.map(get_free_energy, self.sequences)
        self.boxplot(ax, deltag, "Free energy", "deltaG")
        self.logger.debug("Added free energy to axis.")

    def add_off_targets(self, ax: plt.Axes):
        """Add unique number of off targets to axis."""
        task = AlignProbeCandidates()
        task.fname_fasta = RunConfig().analyze_probeset
        task.fname_sam = os.path.splitext(task.fname_fasta)[0] + ".sam"
        task.fname_genome = util.get_genome_name()
        task.fname_gene = util.get_gene_name()

        task.align_probes(
            max_off_targets=20,
            is_endogenous=SequenceConfig().is_endogenous,
            threads=GeneralConfig().threads,
        )
        df_sam = task.parse_raw_pysam(
            pysam.view(task.fname_sam), SequenceConfig().is_endogenous  # type: ignore
        )
        counts = df_sam.groupby("qname").size().values - 1
        self.histplot(ax, counts, "Off target count", -1)

        os.remove(task.fname_sam)
        os.remove(task.fname_fasta.rstrip("a") + "q")
        self.logger.debug("Added off targets to axis.")

    # TODO refactor to merge with optimization.is_binding
    def _get_similarity_score(self, seq1: Bio.Seq.Seq, seq2: Bio.Seq.Seq) -> float:
        alignments = [
            *Bio.pairwise2.align.globalxx(seq1, seq2.complement()),
            *Bio.pairwise2.align.globalxx(seq1, seq2.reverse_complement()),
        ]
        if not alignments:
            return 0.0
        max_score = max(map(lambda x: x.score, alignments))  # type: ignore
        return float(max_score / len(seq1))

    def add_binding_affinity(self, ax: plt.Axes, fig):
        """Add probe vs probe similarity scores (binding) to axis."""
        if len(self.sequences) >= 25:
            eta = len(self.sequences) ** 2 * 20 / 60_000
            self.logger.warning(
                f"Relatively large number of sequences ({len(self.sequences)}). "
                f"Determining sequence similarity will take ~{eta:.2f}mins."
            )
        matrix = np.array(
            [
                [self._get_similarity_score(x.seq, y.seq) for x in self.sequences]
                for y in self.sequences
            ]
        )  # type: ignore
        self.matrix(ax, matrix, "Binding affinity", fig)
        self.logger.debug("Added binding affinity to axis.")

    def add_probe_coverage(self, ax: plt.Axes):
        """Add coverage along gene to axis."""
        coverage = np.zeros(len(self.gene))

        for probe in self.sequences:
            idx = self.gene.seq.find(str(probe.seq))
            coverage[idx : idx + len(probe)] = 1

        ax.set_title("Gene coverage")
        ax.bar(range(len(self.gene)), coverage)
        ax.set(ylabel="Assigned", xlabel="Position")
        self.logger.debug("Added probe coverage to axis.")

    def build_figure(self):
        """Layout to build the figure."""
        fig = plt.figure(figsize=(15, 18))
        plt.suptitle(
            f"Analysis of probe set - {util.get_gene_name(False)}", y=1, fontsize=20
        )
        shape = (4, 3)

        # Row 1
        ax = plt.subplot2grid(shape, (0, 0))
        self.add_length(ax)
        ax = plt.subplot2grid(shape, (0, 1))
        self.add_melting_temperature(ax)
        ax = plt.subplot2grid(shape, (0, 2))
        self.add_gc_content(ax)

        # Row 2
        ax = plt.subplot2grid(shape, (1, 0))
        self.add_g_quadruplet(ax)
        ax = plt.subplot2grid(shape, (1, 1))
        self.add_kmers(ax)
        ax = plt.subplot2grid(shape, (1, 2))
        self.add_free_energy(ax)

        # Row 3
        ax = plt.subplot2grid(shape, (2, 0))
        self.add_off_targets(ax)
        ax = plt.subplot2grid(shape, (2, 1))
        self.add_binding_affinity(ax, fig)

        # TODO add a way to display off target expression percentages?
        # if ProbeConfig().encode_count_table:
        #     ax = plt.subplot2grid(shape, (2, 2))
        #     self.add_expression_percentages(ax)

        # Row 4
        ax = plt.subplot2grid(shape, (3, 0), colspan=3)
        self.add_probe_coverage(ax)

        plt.tight_layout()
        plt.savefig(self.output().path)

    def run(self):
        self.sequences = list(
            Bio.SeqIO.parse(RunConfig().analyze_probeset, format="fasta")
        )
        self.build_figure()
        os.remove(self.input()["sequence"].path)



"""Utility and getter functions."""

from typing import Any, List
import hashlib
import logging
import os
import re
import unicodedata

import Bio.SeqRecord
import luigi
import pandas as pd

from .config import GeneralConfig
from .config import SequenceConfig
from .constants import FASTA_EXT


def get_output_dir(config: luigi.Config = GeneralConfig) -> Any:
    """Return the output directory."""
    output_dir = config().output_dir
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not output_dir:
        output_dir = os.getcwd()
    return os.path.abspath(output_dir)


def hash_fn(key: str) -> str:
    """Deterministic hash function."""
    return hashlib.sha256(key.encode("utf-8")).hexdigest()[:10]


def create_config_hash(
    luigi_config: luigi.configuration.cfg_parser.LuigiConfigParser,
) -> str:
    """Create a hash from a luigi config parser."""
    return hash_fn(
        str(
            {
                section_name: dict(luigi_config[section_name])
                for section_name in luigi_config.sections()
            }
        )
    )


def get_gene_name(hashed: bool = True, config: luigi.Config = SequenceConfig) -> str:
    """Return the gene name without extension."""
    # Name based on gene fasta file
    sequence_file = config().sequence_file
    if sequence_file and os.path.isfile(sequence_file):
        basename = secure_filename(os.path.splitext(os.path.basename(sequence_file))[0])
    # Using Ensembl ID
    elif config().ensembl_id:
        basename = secure_filename(config().ensembl_id)
    # Name based on NCBI parameters
    elif config().gene_name and config().organism_name:
        basename = secure_filename(f"{config().organism_name}_{config().gene_name}")
    else:
        raise ValueError(
            "Could not determine gene name. "
            "Please specify the gene name and organism, or Ensembl ID, or pass in a sequence file."
        )

    if not hashed:
        return basename
    return f"{basename}_{create_config_hash(luigi.configuration.get_config())}"


def get_genome_name(config: luigi.Config = GeneralConfig) -> Any:
    """Return the genome name without extension."""
    if config().reference_genome is None:
        raise ValueError("Reference genome must be passed.")

    if not (
        os.path.exists(config().reference_genome)
        or os.path.splitext(config().reference_genome)[1] in FASTA_EXT
    ):
        raise ValueError("The passed reference genome must be a valid .fa file.")

    return os.path.abspath(os.path.splitext(config().reference_genome)[0])


def secure_filename(filename: str) -> str:
    """Pass a filename and return a secure version of it.

    This filename can then safely be stored on a regular file system and passed
    to `os.path.join`.  The filename returned is an ASCII only string for
    maximum portability.

    On windows system the function also makes sure that the file is not
    named after one of the special device files.
    """
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.encode("ascii", "ignore").decode("ascii")

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    _filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip(
        "._"
    )
    return filename


def log_and_check_candidates(
    logger: logging.Logger, name: str, count: int, count_prev: int = 0
) -> None:
    """Log candidate count before/after filtering."""
    previous = f" (from {count_prev})" if count_prev else ""
    logger.info(f"Writing {count}{previous} candidates in {name}.")

    if count == 0:
        raise ValueError(
            f"No more probes remaining after {name}. "
            "Please check your gene sequence and the stringency of your parameters."
        )
    if count < 10:
        logger.warning(
            f"Only {count} candidates remain after {name}. "
            "Possibly check the stringency of your parameters."
        )


def create_data_table(sequences: List[Bio.SeqRecord.SeqRecord]) -> pd.DataFrame:
    """Create a data table from a list of sequences."""
    df = pd.DataFrame(
        list(map(lambda x: (x.id, len(x)), sequences)), columns=["name", "length"]
    )
    df.insert(1, "sequence", [str(seq.seq) for seq in sequences])
    df["start"] = (
        df["name"].str.extract(r"^candidate-\d+-(\d+)$", expand=True).astype(int)
    )
    df["end"] = df["start"] + df["length"]
    df = df.sort_values(by="start", ignore_index=True)
    return df


class UniCode:
    """Addition of fancy colors in help text."""

    ispos = os.name == "posix"
    blue = "\033[34m" if ispos else ""
    bold = "\033[1m" if ispos else ""
    cyan = "\033[36m" if ispos else ""
    green = "\033[32m" if ispos else ""
    magenta = "\033[35m" if ispos else ""
    red = "\033[31m" if ispos else ""
    yellow = "\033[33m" if ispos else ""
    end = "\033[0m" if ispos else ""
    dash = "\U0001F4A8"
    dna = "\U0001F9EC"
    error = "\U0001f6d1"
    fishing = "\U0001F3A3"
    happy = "\U0001F603"
    party = "\U0001F973"
    warn = "\u26A0"



"""Create a list of candidate probes from a gene sequence."""

import logging
import os
from typing import List

import Bio.SeqIO
import Bio.SeqRecord
import Bio.SeqUtils
import Bio.SeqUtils.MeltingTemp
import luigi

from . import util
from .config import ProbeConfig
from .prepare_sequence import PrepareSequence


class GenerateAllProbes(luigi.Task):
    """Create all possible probes in a gene given length ranges."""

    logger = logging.getLogger("luigi-interface")

    def requires(self):
        return PrepareSequence()

    def output(self):
        return luigi.LocalTarget(
            os.path.join(util.get_output_dir(), f"{util.get_gene_name()}_all.fasta")
        )

    def create_candidate_probes(
        self, sequence: Bio.SeqRecord.SeqRecord, min_length: int, max_length: int
    ) -> List[Bio.SeqRecord.SeqRecord]:
        """Create a set of all subsequences of sequence with the right lengths."""
        if min_length > max_length:
            raise ValueError(
                "Minimum probe length must be smaller or equal to maximum length. "
                f"{min_length} > {max_length}!"
            )
        if min_length >= len(sequence):
            raise ValueError(
                "Minimum probe length must be shorter than the sequence length. "
                f"{min_length} >= {len(sequence)}!"
            )

        candidates = []
        idx = 1
        for length in range(min_length, max_length + 1):
            for start_pos in range(0, len(sequence) - length + 1):
                candidates.append(
                    Bio.SeqRecord.SeqRecord(
                        sequence.seq[start_pos : start_pos + length],
                        id=f"candidate-{idx}-{start_pos}",
                    )
                )
                idx += 1
        return candidates

    def run(self):
        sequence = Bio.SeqIO.read(self.input().path, format="fasta")
        candidates = self.create_candidate_probes(
            sequence, ProbeConfig().min_length, ProbeConfig().max_length
        )
        util.log_and_check_candidates(self.logger, "GenerateAllProbes", len(candidates))
        Bio.SeqIO.write(candidates, self.output().path, format="fasta")



"""Predict secondary structure of probes and filter them based on deltaG."""

from pathlib import Path
import logging
import multiprocessing
import os
import re
import subprocess
import sys
import tempfile

import Bio.SeqIO
import Bio.SeqRecord
import luigi

from . import util
from .config import GeneralConfig
from .config import ProbeConfig
from .kmers import KMerFiltering


def get_free_energy(sequence: Bio.SeqRecord.SeqRecord) -> float:
    """Return the predicted free energy of the sequence.

    Using "Fold" as part of the RNAstructure package from the Mathews lab.
    """
    file_path = Path(__file__).resolve().parent.as_posix()
    data_table = os.path.join(file_path, "data_tables/")
    if sys.platform == "linux" or sys.platform == "linux2":
        fold_path = "Fold"
    if sys.platform == "darwin":
        fold_path = os.path.join(file_path, "Fold_osx")

    os.environ["DATAPATH"] = data_table

    with tempfile.TemporaryDirectory() as tmp_dir:
        fname = os.path.join(tmp_dir, "input.fasta")
        Bio.SeqIO.write(sequence, fname, format="fasta")
        args_fold = [fold_path, fname, "-", "--bracket", "--MFE"]
        sec = subprocess.check_output(args_fold, stderr=subprocess.STDOUT).decode()

    # Format with secondary structure:
    # >ENERGY = -deltaG NAME DESCRIPTION\nSEQENCE\nDOT-BRACKET
    # Format without secondary structure:
    # >NAME DESCRIPTION\nSEQUENCE\n...
    search = re.search(r">ENERGY = (-\d+\.\d)\s\s([\w-]+)", sec)
    if search is not None:
        energy = float(search.groups()[0])
    else:
        energy = 0.0
    return energy


class SecondaryStructureFiltering(luigi.Task):
    """Predict free energy of secondary structures and filter them based on deltaG."""

    logger = logging.getLogger("custom-logger")

    def requires(self):
        return KMerFiltering()

    def output(self):
        return luigi.LocalTarget(
            os.path.join(util.get_output_dir(), f"{util.get_gene_name()}_sstruct.fasta")
        )

    def run(self):
        sequences = list(Bio.SeqIO.parse(self.input().path, format="fasta"))

        with multiprocessing.Pool(GeneralConfig().threads) as pool:
            free_energies = pool.map(get_free_energy, sequences)

        # Filter out candidates with low free energy
        candidates = [
            sequence
            for sequence, energy in zip(sequences, free_energies)
            if energy >= ProbeConfig().max_deltag
        ]

        util.log_and_check_candidates(
            self.logger, "SecondaryStructureFiltering", len(candidates), len(sequences)
        )
        Bio.SeqIO.write(candidates, self.output().path, format="fasta")



"""eFISHent."""

__version__ = "0.0.2"

import logging

# Silence some troublemakers
logging.getLogger("luigi-interface").setLevel(level=logging.CRITICAL)
logging.getLogger("matplotlib.font_manager").setLevel(level=logging.CRITICAL)

from . import alignment
from . import analyze
from . import basic_filtering
from . import cleanup
from . import cli
from . import config
from . import constants
from . import generate_probes
from . import kmers
from . import optimization
from . import prepare_sequence
from . import secondary_structure
from . import util



"""Find the optimal probeset with the highest coverage (fast or optimal)."""

from typing import List, Tuple
import functools
import logging
import multiprocessing
import os

import Bio.Seq
import Bio.SeqIO
import Bio.pairwise2
import luigi
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyomo.environ as pe

from . import util
from .config import GeneralConfig
from .config import ProbeConfig
from .config import RunConfig
from .secondary_structure import SecondaryStructureFiltering


class OptimizeProbeCoverage(luigi.Task):
    """Find the optimal probeset with the highest coverage (fast or optimal)."""

    logger = logging.getLogger("custom-logger")

    def requires(self):
        return SecondaryStructureFiltering()

    def output(self):
        return {
            name: luigi.LocalTarget(os.path.join(util.get_output_dir(), filename))
            for name, filename in [
                ("probes", f"{util.get_gene_name()}_optimal.fasta"),
                ("coverage", f"{util.get_gene_name()}_optimization.png"),
            ]
        }

    def run_optimal(self, threads: int, time_limit: int):
        """Run the optimal model by parallelizing non-overlapping segments."""
        index_block = 0
        prev_probe = self.df.iloc[0]
        blocks = []

        for _, probe in self.df.iterrows():
            if not is_overlapping(
                prev_probe[["start", "end"]], probe[["start", "end"]]
            ):
                index_block += 1
            prev_probe = probe
            blocks.append(index_block)
        self.df["block"] = blocks

        self.time_limit = time_limit
        with multiprocessing.Pool(threads) as pool:
            _assigned = pool.map(self.run_optimal_block, self.df["block"].unique())

        # Flatten the list of lists
        assigned = functools.reduce(lambda x, y: x + y, _assigned)
        return assigned

    def run_optimal_block(self, block: int) -> List[str]:
        """Run a single non-overlapping block."""
        df = self.df[self.df["block"] == block]
        return optimal_model(df, self.time_limit)

    def run_greedy(self):
        """Run sequential greedy model."""
        return greedy_model(self.df)

    def filter_binding_probes(
        self, assigned: List[str], match_percentage: float
    ) -> List[str]:
        """Remove probes of too high rev/comp sequence similarity.

        It would be cleaner to have it evaluated at run time / not have them included
        to begin with but that'd be O(n2) with potentially 1k+ probes leading to a
        runtime of multiple hours (with 20ms per `is_binding` call).
        """
        assigned_sequences = self.df.loc[
            self.df["name"].isin(assigned), "sequence"
        ].values

        # Create probe vs probe matrix that could bind
        binding_matrix = np.array(
            [
                [is_binding(x, y, match_percentage) for x in assigned_sequences]
                for y in assigned_sequences
            ]
        )  # type: ignore

        trouble_makers = np.unique(np.where(binding_matrix)[1])  # type: ignore
        self.logger.debug(f"Found {len(trouble_makers)} trouble makers to be removed.")
        assigned = [
            name for idx, name in enumerate(assigned) if idx not in trouble_makers
        ]
        return assigned

    def run(self):
        sequences = list(Bio.SeqIO.parse(self.input().path, "fasta"))
        self.df = util.create_data_table(sequences)
        self.df["end"] += ProbeConfig().spacing

        if RunConfig().optimization_method == "greedy":
            assigned = self.run_greedy()
        elif RunConfig().optimization_method == "optimal":
            assigned = self.run_optimal(
                threads=GeneralConfig().threads,
                time_limit=RunConfig().optimization_time_limit,
            )
        else:
            raise ValueError(
                f"Invalid optimization method: {RunConfig().optimization_method}"
                "Must be greedy or optimal."
            )
        if ProbeConfig().sequence_similarity > 0:
            match_percentage = ProbeConfig().sequence_similarity / 100
            assigned = self.filter_binding_probes(assigned, match_percentage)

        visualize_assignment(self.df, assigned, self.output()["coverage"].path)

        candidates = [sequence for sequence in sequences if sequence.id in assigned]
        util.log_and_check_candidates(
            self.logger, "OptimizeProbeCoverage", len(candidates), len(sequences)
        )
        Bio.SeqIO.write(candidates, self.output()["probes"].path, format="fasta")


def is_overlapping(x: Tuple[int, int], y: Tuple[int, int]) -> bool:
    """Check if two ranges overlap."""
    x_set = set(range(x[0], x[1] + 1))
    return len(x_set.intersection(range(y[0], y[1] + 1))) != 0


def is_binding(seq1: str, seq2: str, match_percentage: float = 0.75) -> bool:
    """Check if seq1 is similar to (rev) complement of seq2 / if would bind."""
    if 0 > match_percentage > 1:
        raise ValueError(
            "Matching percentage must be between 0 and 100. "
            f"Found `{match_percentage * 100}`"
        )
    alignments = [
        *Bio.pairwise2.align.globalxx(seq1, Bio.Seq.Seq(seq2).complement()),
        *Bio.pairwise2.align.globalxx(seq1, Bio.Seq.Seq(seq2).reverse_complement()),
    ]
    if not alignments:
        return False

    max_score = max(map(lambda x: x.score, alignments))  # type: ignore
    if (max_score / len(seq1)) <= match_percentage:
        return False

    return True


def greedy_model(df: pd.DataFrame) -> List[str]:
    """Run the greedy/fast model."""
    rec_array = df[["start", "end"]].to_records(index=False)
    probes = df["name"].values

    # Create overlap matrix of all vs all
    vect = np.vectorize(is_overlapping)
    matrix = vect(rec_array[:, None], rec_array)
    overlap = {probe: row for probe, row in zip(probes, matrix)}

    # Assign non-ovelapping probes to the first probe
    assign = {}
    assigned = [probes[0]]
    for idx, probe in enumerate(probes):
        if not overlap[assigned[-1]][idx]:
            assigned.append(probe)
        assign[probe] = probe in assigned
    return assigned


# TODO add proper mathematical description
def optimal_model(df: pd.DataFrame, time_limit: int) -> List[str]:
    """Run the optimal mathematical model."""
    # Convert dataframe to usable model inputs
    sequence = list(range(df["start"].min(), df["end"].max()))
    probes = df["name"].values
    probe_starts = {k: v for _, (k, v) in df[["name", "start"]].iterrows()}
    probe_ends = {k: v for _, (k, v) in df[["name", "end"]].iterrows()}

    # Model to make non-contiguous connections across a sequence
    # with objective to "cover" as many points in sequence as possible
    coverages = {
        p: [t for t in sequence if t >= probe_starts[p] and t <= probe_ends[p]]
        for p in probes
    }

    # Model definition
    model = pe.ConcreteModel()
    model.sequence = pe.Set(initialize=sequence)
    model.probes = pe.Set(initialize=probes)
    model.covers = pe.Set(model.probes, within=model.sequence, initialize=coverages)
    model.covers_flat_set = pe.Set(
        initialize=[(p, s) for p in probes for s in model.covers[p]]
    )
    model.assign = pe.Var(model.probes, domain=pe.Binary)
    model.covered = pe.Var(model.covers_flat_set, domain=pe.Binary)

    # Objectiv
    obj = sum(model.covered[p, s] for (p, s) in model.covers_flat_set)
    model.objective = pe.Objective(expr=obj, sense=pe.maximize)

    # Constraints
    # Selected probe must cover the associated points between start and end, if assigned
    def cover(model, p):
        return (
            sum(model.covered[p, s] for s in model.covers[p])
            == len(model.covers[p]) * model.assign[p]
        )

    # Cannot cover any point by more than 1 probe
    def over_cover(model, s):
        cov_options = [(p, s) for p in model.probes if (p, s) in model.covers_flat_set]
        if not cov_options:
            return pe.Constraint.Skip  # no possible coverages
        return sum(model.covered[p, s] for (p, s) in cov_options) <= 1

    model.C1 = pe.Constraint(model.probes, rule=cover)
    model.C2 = pe.Constraint(model.sequence, rule=over_cover)
    solver = pe.SolverFactory("glpk")
    solver.options["tmlim"] = time_limit
    solver.solve(model, tee=False)

    assigned = [name for name, assign in model.assign.get_values().items() if assign]
    return assigned


def visualize_assignment(df: pd.DataFrame, assigned: List[str], filename: str) -> None:
    """Visualize the assignment as overlap matrix."""
    start = df["start"].min()
    end = df["end"].max()
    matrix = np.zeros((len(df) + 1, end - start))
    for idx, (_, row) in enumerate(df.iterrows()):
        matrix[idx + 1, row["start"] - start : row["end"] - start + 1] = (
            2 if row["name"] in assigned else 1
        )

    plt.figure()
    plt.title(f"Probe Coverage from {start} to {end}")
    plt.imshow(matrix)
    plt.tight_layout()
    plt.savefig(filename, dpi=600, bbox_inches="tight")



"""Download Entrez Gene Probes from NCBI.

Select the right strand and select intronic/exonic regions.
"""

from typing import List
import logging
import os
import subprocess

import Bio.SeqIO
import Bio.SeqRecord
import luigi

from . import util
from .config import SequenceConfig


class DownloadEntrezGeneSequence(luigi.Task):
    """Download genomic sequence as fasta from NCBI."""

    logger = logging.getLogger("custom-logger")

    def output(self):
        fname = f"{util.get_gene_name()}_entrez.fasta"
        return luigi.LocalTarget(os.path.join(util.get_output_dir(), fname))

    def get_entrez_query(
        self, ensembl_id: str, gene_name: str, organism_name: str
    ) -> str:
        """Retrieve the query depending on sequence configuration arguments passed."""
        if ensembl_id:
            return f"({ensembl_id})" + (
                f" AND {organism_name}[Organism]" if organism_name else ""
            )

        if gene_name and organism_name:
            return f"({gene_name}[Gene Name]) AND {organism_name}[Organism]"

        raise ValueError(
            "For downloading Entrez Gene Probes, "
            " you need to specify the gene name and organism name or provide an Emsembl ID."
        )

    def fetch_entrez(self, query: str) -> str:
        """Retrieve the fasta sequence data from entrez."""
        args_search = ["esearch", "-db", "gene", "-query", query]
        args_link = [
            "elink",
            "-db",
            "gene",
            "-target",
            "nuccore",
            "-name",
            "gene_nuccore_refseqrna",
        ]
        args_fetch = ["efetch", "-format", "fasta"]
        self.logger.debug(f"Fetching from Entrez using query '{query}'.")

        search = subprocess.run(args_search, check=True, capture_output=True)
        link = subprocess.run(args_link, input=search.stdout, capture_output=True)
        fetch = subprocess.run(args_fetch, input=link.stdout, capture_output=True)
        fasta = fetch.stdout.decode()

        if not fasta:
            raise LookupError(
                "No Entrez gene found. "
                "Please check your gene name and organism and try again. "
                "Or provide a custom fasta file as `sequence-file`."
            )
        return fasta

    def run(self):
        query = self.get_entrez_query(
            SequenceConfig().ensembl_id,
            SequenceConfig().gene_name,
            SequenceConfig().organism_name,
        )
        fasta = self.fetch_entrez(query)

        with self.output().open("w") as outfile:
            outfile.write(fasta)
        self.logger.info(f"Downloaded sequence from entrez using query '{query}'.")


class PrepareSequence(luigi.Task):
    """Prepare gene sequence.

    Select right strand and exon/intron/both.
    """

    logger = logging.getLogger("custom-logger")

    def requires(self):
        tasks = {}
        if not SequenceConfig().sequence_file:
            tasks["entrez"] = DownloadEntrezGeneSequence()
        return tasks

    def output(self):
        fname = f"{util.get_gene_name()}_sequence.fasta"
        return luigi.LocalTarget(os.path.join(util.get_output_dir(), fname))

    def select_sequence(
        self, sequences: List[Bio.SeqRecord.SeqRecord]
    ) -> Bio.SeqRecord.SeqRecord:
        """Select a single sequence as template."""
        if not sequences:
            raise ValueError(
                "No records found in fasta file. "
                "Please ensure at least one sequence is present."
            )
        if len(sequences) > 1:
            self.logger.warning(
                f"{util.UniCode.warn} More than one record (potential isoforms) found in fasta file. "
                "Will default to take the first sequence."
            )
        return sequences[0]

    def select_strand(
        self, sequence: Bio.SeqRecord.SeqRecord, is_plus_strand: bool
    ) -> Bio.SeqRecord.SeqRecord:
        """Select right strand for probes to target."""
        if is_plus_strand:
            sequence = sequence.reverse_complement(id=True, description=True)
            self.logger.debug("Converted sequence to reverse complement.")
        return sequence

    def run(self):
        input_file = (
            self.input()["entrez"].path
            if "entrez" in self.input()
            else SequenceConfig().sequence_file
        )
        sequences = list(Bio.SeqIO.parse(input_file, format="fasta"))
        sequence = self.select_sequence(sequences)
        sequence = self.select_strand(sequence, SequenceConfig().is_plus_strand)
        self.logger.info(f'Selected sequence and strand called "{sequence.id}".')
        Bio.SeqIO.write(sequence, self.output().path, format="fasta")



"""Common constants used in other methods."""
from typing import List
import luigi

from .config import GeneralConfig
from .config import ProbeConfig
from .config import RunConfig
from .config import SequenceConfig

SAMFILE_COLUMNS = [
    "qname",
    "flag",
    "rname",
    "pos",
    "mapq",
    "cigar",
    "rnext",
    "pnext",
    "tlen",
    "seq",
    "qual",
    "XA",
    "MD",
    "NM",
]

BLAST_COLUMNS = [
    "qseqid",
    "sseqid",
    "pident",
    "length",
    "mismatch",
    "gapopen",
    "qstart",
    "qend ",
    "sstart",
    "send",
    "evalue",
    "bitscore",
]

GTF_COLUMNS = [
    "gene_id",
    "seqname",
    "start",
    "end",
    "frame",
    "exon_number",
    "feature",
]

CONFIG_CLASSES: List[luigi.Config] = [
    GeneralConfig,
    RunConfig,
    SequenceConfig,
    ProbeConfig,
]

FASTA_EXT = (".fasta", ".fna", ".ffn", ".faa", ".frn", ".fa")

SAM_FLAG_REVERSE = "16"

CLI_SHORTFORM = {
    "analyze_probeset": "ap",
    "build_indices": "idx",
    "encode_count_table": "ct",
    "ensembl_id": "id",
    "formamide_concentration": "f",
    "gene_name": "gen",
    "is_endogenous": "endo",
    "is_plus_strand": "plus",
    "kmer_length": "kl",
    "max_deltag": "dg",
    "max_expression_percentage": "ep",
    "max_gc": "GC",
    "max_kmers": "km",
    "max_length": "L",
    "max_off_targets": "ot",
    "max_tm": "TM",
    "min_gc": "gc",
    "min_length": "l",
    "min_tm": "tm",
    "na_concentration": "na",
    "optimization_method": "om",
    "optimization_time_limit": "otl",
    "organism_name": "org",
    "output_dir": "o",
    "reference_annotation": "a",
    "reference_genome": "g",
    "save_intermediates": "int",
    "sequence_file": "seq",
    "sequence_similarity": "sim",
    "spacing": "sp",
    "threads": "t",
}



"""Filter probes based on simple criteria."""

from typing import List
import logging
import multiprocessing
import os

import Bio.Seq
import Bio.SeqIO
import Bio.SeqRecord
import Bio.SeqUtils
import Bio.SeqUtils.MeltingTemp
import luigi

from . import util
from .config import GeneralConfig
from .config import ProbeConfig
from .generate_probes import GenerateAllProbes


def get_melting_temp(
    sequence: Bio.Seq.Seq, na_concentration: float, formamide_concentration: float
) -> float:
    """Get melting temperature of candidate."""
    tm_raw = Bio.SeqUtils.MeltingTemp.Tm_NN(sequence, Na=na_concentration)
    melting_temp = Bio.SeqUtils.MeltingTemp.chem_correction(
        tm_raw, fmd=formamide_concentration
    )
    return float(melting_temp)


def get_gc_content(sequence: Bio.Seq.Seq) -> float:
    """Get GC content of candidate."""
    return float(Bio.SeqUtils.GC(sequence))


def get_g_quadruplet_count(sequence: Bio.Seq.Seq) -> int:
    """Get number of G quadruplets in candidate."""
    return int(sequence.count("GGGG"))


class BasicFiltering(luigi.Task):
    """Initial probe filtering based on melting temperature and GC content."""

    logger = logging.getLogger("custom-logger")

    def requires(self):
        return GenerateAllProbes()

    def output(self):
        return luigi.LocalTarget(
            os.path.join(util.get_output_dir(), f"{util.get_gene_name()}_basic.fasta")
        )

    def is_candidate_valid(
        self, candidate: Bio.SeqRecord.SeqRecord, config: luigi.Config
    ) -> bool:
        """Check if candidate matches basic criteria."""
        if config.min_tm > config.max_tm:
            raise ValueError("Max TM has to be greater or equal the min TM!")
        if config.min_gc > config.max_gc:
            raise ValueError("Max GC has to be greater or equal the min GC!")

        sequence = candidate.seq
        gc_content = get_gc_content(sequence)
        g_quadruplets = get_g_quadruplet_count(sequence)
        melting_temp = get_melting_temp(
            sequence, config.na_concentration, config.formamide_concentration
        )

        if (
            config.min_gc <= gc_content <= config.max_gc
            and config.min_tm <= melting_temp <= config.max_tm
            and g_quadruplets == 0
        ):
            return True
        return False

    def _is_candidate_valid(self, candidate: Bio.SeqRecord.SeqRecord) -> bool:
        """Counteracting the really weird multiprocess class behavior."""
        return self.is_candidate_valid(candidate, self.config)

    def screen_sequences(
        self, sequences: List[Bio.SeqRecord.SeqRecord]
    ) -> List[Bio.SeqRecord.SeqRecord]:
        """Filter candiates not matching criteria."""
        if len(sequences) > 1000:
            self.config = ProbeConfig()
            with multiprocessing.Pool(GeneralConfig().threads) as pool:
                valid = pool.map(self._is_candidate_valid, sequences)
        else:
            valid = [
                self.is_candidate_valid(candidate, ProbeConfig())
                for candidate in sequences
            ]

        candidates = [seq for seq, valid in zip(sequences, valid) if valid]
        return candidates

    def run(self):
        sequences = list(Bio.SeqIO.parse(self.input().path, "fasta"))
        candidates = self.screen_sequences(sequences)
        util.log_and_check_candidates(
            self.logger, "BasicFiltering", len(candidates), len(sequences)
        )
        Bio.SeqIO.write(candidates, self.output().path, format="fasta")



"""Create jellyfish indices of the reference genome.

Filter probes based on jellyfish indices.
"""

import logging
import multiprocessing
import os
import subprocess

import Bio.SeqIO
import Bio.SeqRecord
import luigi

from . import util
from .alignment import AlignProbeCandidates
from .config import GeneralConfig
from .config import ProbeConfig


def get_max_kmer_count(sequence: Bio.SeqRecord, jellyfish_path: str) -> int:
    """Count kmers in a sequence.

    Only keep the sequence if it has less than max_kmers kmers.
    """
    if len(sequence.seq) < ProbeConfig().kmer_length:
        raise ValueError(
            f"Probe length must be larger than kmer length ({ProbeConfig().kmer_length})."
        )

    sub_kmers = [
        str(sequence.seq[i : i + ProbeConfig().kmer_length])
        for i in range(len(sequence) - ProbeConfig().kmer_length + 1)
    ]
    args_jellyfish = ["jellyfish", "query", jellyfish_path, " ".join(sub_kmers)]
    kmer_counts_raw = subprocess.check_output(
        args_jellyfish, stderr=subprocess.STDOUT
    ).decode()

    # Format of jellyfish output: "kmer1 count1\nkmer2 count2\n..."
    kmer_counts = list(
        map(lambda x: int(x.split(" ")[1]), kmer_counts_raw.split("\n")[:-1])
    )
    return max(kmer_counts) if kmer_counts else 0


class BuildJellyfishIndex(luigi.Task):
    """Index building task for jellyfish kmers."""

    logger = logging.getLogger("custom-logger")

    def output(self):
        return luigi.LocalTarget(
            f"{util.get_genome_name()}_{ProbeConfig().kmer_length}.jf"
        )

    def run(self):
        if ProbeConfig().max_kmers <= 2:
            self.logger.warning(
                f"{util.UniCode.warn} Jellyfish index will be created but not used because max_kmers <= 2."
            )

        args_jellyfish = [
            "jellyfish",
            "count",
            "--mer-len",
            str(ProbeConfig().kmer_length),
            "--out-counter-len",
            "1",
            "--lower-count",
            "2",
            "--size",
            "100M",
            "--threads",
            str(GeneralConfig().threads),
            "--output",
            self.output().path,
            GeneralConfig().reference_genome,
        ]
        self.logger.debug(f"Running jellyfish with - {' '.join(args_jellyfish)}")
        subprocess.check_call(args_jellyfish)
        self.logger.info(f"Finished building kmer ({ProbeConfig().kmer_length}) index.")


class KMerFiltering(luigi.Task):
    """Filter probes containing too many common short kmers."""

    logger = logging.getLogger("custom-logger")

    def requires(self):
        return {"jellyfish": BuildJellyfishIndex(), "probes": AlignProbeCandidates()}

    def output(self):
        return luigi.LocalTarget(
            os.path.join(util.get_output_dir(), f"{util.get_gene_name()}_kmer.fasta")
        )

    def _get_kmer_count(self, sequence: Bio.SeqRecord) -> int:
        return get_max_kmer_count(sequence, self.path)

    def run(self):
        sequences = list(
            Bio.SeqIO.parse(self.input()["probes"]["fasta"].path, format="fasta")
        )

        # TODO WTF... if this is passed as reference and not as variable
        # multiprocessing will fk things up so that GeneralConfig() is reset...
        self.path = self.input()["jellyfish"].path

        with multiprocessing.Pool(GeneralConfig().threads) as pool:
            counts = pool.map(self._get_kmer_count, sequences)

        candidates = [
            seq
            for seq, count in zip(sequences, counts)
            if count <= ProbeConfig().max_kmers
        ]

        util.log_and_check_candidates(
            self.logger, "KMerFiltering", len(candidates), len(sequences)
        )
        Bio.SeqIO.write(candidates, self.output().path, format="fasta")



"""Command line interface."""

from pathlib import Path
from typing import Any, List
import argparse
import configparser
import logging
import os
import sys
import tempfile

import luigi

from .analyze import AnalyzeProbeset
from .cleanup import CleanUpOutput
from .constants import CLI_SHORTFORM
from .constants import CONFIG_CLASSES
from .indexing import BuildBlastDatabase
from .indexing import BuildBowtieIndex
from .kmers import BuildJellyfishIndex
from .util import UniCode


GROUP_DESCRIPTIONS = {
    f"{UniCode.blue} General": "General configuration that will be used for all tasks.",
    f"{UniCode.green} Run": "Options that change the behavior of the workflow.",
    f"{UniCode.red} Sequence": "Details about the sequences the probe design will be performed on.",
    f"{UniCode.magenta} Probe": "Probe filtering and design options.",
}
REQUIRED_PARAMS = ["reference_genome"]


def string_to_bool(value) -> bool:
    """Workaround for using typed boolean values as arguments."""
    if isinstance(value, bool):
        return value
    if value.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif value.lower() in ("no", "false", "f", "n", "0"):
        return False

    raise argparse.ArgumentTypeError("Boolean value expected.")


def get_parameter_type(param: luigi.Parameter) -> Any:
    """Get the type of a parameter."""
    if isinstance(param, luigi.IntParameter):
        return int
    elif isinstance(param, luigi.FloatParameter):
        return float
    elif isinstance(param, luigi.BoolParameter):
        return string_to_bool
    return str


def _add_utilities(parser: argparse.ArgumentParser) -> None:
    """Add the utility arguments to the parser."""
    utility = parser.add_argument_group(f"{UniCode.cyan}General utilities{UniCode.end}")
    utility.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this message.",
    )
    utility.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s 0.0.1",
        help="Show %(prog)s's version number.",
    )
    utility.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help="Change the program output to silent hiding information on progress.",
    )
    utility.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)


def _add_group(group: argparse._ArgumentGroup, config_class: luigi.Config) -> None:
    """Add a single configuration class/group to a parser."""
    for name, param in config_class().get_params():
        param_type = get_parameter_type(param)
        is_required = name in REQUIRED_PARAMS
        default = (
            "-"
            if (
                param._default is None
                or param._default == ""
                and param_type != string_to_bool
            )
            else param._default
        )
        group.add_argument(
            f"-{CLI_SHORTFORM.get(name)}",
            f"--{name.replace('_', '-')}",
            type=param_type,
            required=is_required,
            default=param._default,
            help=f"{param.description} "
            f"[default: {default}, "
            f"required: {is_required}]",
        )


def _add_groups(parser: argparse.ArgumentParser) -> None:
    """Add the main option groups to the parser."""
    groups = [
        parser.add_argument_group(
            f"{name} options{UniCode.end}", description=description
        )
        for name, description in GROUP_DESCRIPTIONS.items()
    ]

    for group, config_class in zip(groups, CONFIG_CLASSES):
        _add_group(group, config_class)


def _parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        prog="eFISHent",
        description=f"{UniCode.bold}eFISHent {UniCode.fishing} {UniCode.dna} to design all your probes.{UniCode.end}",
        epilog=(
            'See the online wiki at "https://github.com/BBQuercus/eFISHent/wiki" for an overview.\n'
            f"We hope you enjoy using eFISHent {UniCode.party}!"
        ),
        add_help=False,
    )
    _add_groups(parser)
    _add_utilities(parser)
    try:
        if len(sys.argv) == 1:
            parser.print_help()
            parser.exit(0)
    except Exception as e:
        print(e)
    return parser.parse_args()


def create_custom_config(args: argparse.Namespace, config_file: str) -> None:
    """Create a custom config file."""
    config = configparser.ConfigParser()
    config_path = Path(__file__).resolve().parent.joinpath("luigi.cfg").as_posix()
    config.read(config_path)

    for section, config_class in zip(
        ["GeneralConfig", "RunConfig", "SequenceConfig", "ProbeConfig"], CONFIG_CLASSES
    ):
        for name in config_class().get_param_names():
            value = vars(args).get(name)
            config.set(section, name, str(value))
            if name == "threads":
                threads = min(vars(args).get(name), os.cpu_count())  # type: ignore
                config.set(section, name, str(threads))

    with open(config_file, "w") as f:
        config.write(f)


def set_logging_level(silent: bool, debug: bool) -> logging.Logger:
    """Set the logging level of luigi and custom logger."""
    log_format = "%(asctime)s %(levelname)-4s - %(message)s"
    luigi_level = "WARNING"
    logfile = None

    if debug:
        log_format = (
            "%(asctime)s %(levelname)-4s [%(name)s] "
            "%(filename)s %(funcName)s %(lineno)d / %(thread)d - %(message)s"
        )
        luigi_level = "DEBUG"
        custom_level = logging.DEBUG
        logfile = "efishent.log"
    elif silent:
        custom_level = logging.WARNING
    else:
        custom_level = logging.INFO

    logging.basicConfig(filename=logfile, format=log_format, force=True)  # type: ignore
    logging.getLogger("luigi").setLevel(luigi_level)
    logging.getLogger("luigi-interface").setLevel(luigi_level)
    luigi.interface.core.log_level = luigi_level

    logger = logging.getLogger("custom-logger")
    logger.setLevel(custom_level)
    return logger


def main():
    """Run eFISHent tasks."""
    args = _parse_args()
    logger = set_logging_level(args.silent, args.debug)
    logger.info(f"{UniCode.fishing} eFISHent has started running...")

    with tempfile.TemporaryDirectory() as tmp_dir:
        config_file = os.path.join(tmp_dir, "luigi.cfg")
        create_custom_config(args, config_file)
        luigi.configuration.add_config_path(config_file)

        tasks: List[luigi.Task] = []
        if args.build_indices:
            tasks = [BuildJellyfishIndex(), BuildBowtieIndex(), BuildBlastDatabase()]
        elif args.analyze_probeset:
            tasks = [AnalyzeProbeset()]
        else:
            tasks = [CleanUpOutput()]

        luigi.build(tasks, local_scheduler=True)

    if tasks[-1].complete():
        logger.info(f"{UniCode.party} eFISHent has finished running!")



"""Clean up the output directory.

Remove the files that are not needed and prettify the kept output.
"""

from typing import List, Optional
import glob
import logging
import os

import luigi
import pandas as pd
import Bio.SeqIO

from . import util
from .alignment import AlignProbeCandidates
from .basic_filtering import get_gc_content
from .basic_filtering import get_melting_temp
from .config import ProbeConfig
from .config import RunConfig
from .config import SequenceConfig
from .kmers import BuildJellyfishIndex
from .kmers import get_max_kmer_count
from .optimization import OptimizeProbeCoverage
from .secondary_structure import get_free_energy


class CleanUpOutput(luigi.Task):
    """Clean up the output files and remove the intermediaries that are not needed."""

    logger = logging.getLogger("custom-logger")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_blast_required = (
            ProbeConfig().encode_count_table and SequenceConfig().is_endogenous
        )

    def requires(self):
        tasks = {
            "optimize": OptimizeProbeCoverage(),
            "jellyfish": BuildJellyfishIndex(),
        }
        if self.is_blast_required:
            tasks["alignment"] = AlignProbeCandidates()
        return tasks

    def output(self):
        return {
            name: luigi.LocalTarget(os.path.join(util.get_output_dir(), filename))
            for name, filename in [
                ("fasta", f"{util.get_gene_name()}.fasta"),
                ("table", f"{util.get_gene_name()}.csv"),
                ("config", f"{util.get_gene_name()}.txt"),
            ]
        }

    def prettify_table(
        self,
        sequences: List[Bio.SeqRecord.SeqRecord],
        basename: str,
        jellyfish_path: str,
        alignment_path: Optional[str] = None,
        config: luigi.Config = ProbeConfig,
    ) -> pd.DataFrame:
        """Create table with probe information."""
        # Create basic table with probe start/end positions and name
        df = util.create_data_table(sequences)

        # Add data columns
        df["GC"] = [round(get_gc_content(seq.seq), 2) for seq in sequences]
        df["TM"] = [
            round(
                get_melting_temp(
                    seq.seq, config().na_concentration, config().formamide_concentration
                ),
                2,
            )
            for seq in sequences
        ]
        df["deltaG"] = [get_free_energy(seq) for seq in sequences]
        df["kmers"] = [get_max_kmer_count(seq, jellyfish_path) for seq in sequences]
        if alignment_path is not None:
            df_counts = pd.read_csv(alignment_path)
            df["count"] = pd.merge(
                df, df_counts, how="left", left_on="name", right_on="qname"
            )["count"].fillna(0)

        # Create new/clean names
        df["name"] = [f"{basename}-{idx + 1}" for idx in df.index]
        return df

    def prettify_sequences(self, df: pd.DataFrame) -> List[Bio.SeqRecord.SeqRecord]:
        """Clean up sequence id's and descriptions."""
        return [
            Bio.SeqRecord.SeqRecord(
                Bio.Seq.Seq(seq), id=name, name=name, description=""
            )
            for seq, name in zip(df["sequence"], df["name"])
        ]

    def prettify_section(self, section: str) -> str:
        """Prettify a single configuration section."""
        pretty = [f"{section.rstrip('Config')} configuration:"]
        for key, value in self.config[section].items():
            if value and value != '""':
                pretty.append(f"  {key.replace('_', '-')}: {value}")
        return "\n".join(pretty)

    def prettify_configuration(self) -> str:
        """Create a pretty file with all set/default configuration."""
        self.config = luigi.configuration.get_config()
        pretty = [self.prettify_section(section) for section in self.config.sections()]
        return "\n\n".join(pretty)

    def remove_intermediates(self) -> None:
        """Remove intermediary files no longer used."""
        intermediary_files = glob.glob(
            os.path.join(util.get_output_dir(), f"{util.get_gene_name()}_*")
        )
        self.logger.debug(f"Found file intermediates - {intermediary_files}")
        try:
            intermediary_files.remove(
                os.path.join(
                    util.get_output_dir(), f"{util.get_gene_name()}_entrez.fasta"
                )
            )
        except ValueError:
            self.logger.debug("Didn't find entrez file...")
        for filename in intermediary_files:
            self.logger.debug(f"Removing {filename}")
            os.remove(filename)

    def run(self):
        sequences = list(
            Bio.SeqIO.parse(self.input()["optimize"]["probes"].path, "fasta")
        )
        df = self.prettify_table(
            sequences,
            basename=util.get_gene_name(hashed=False),
            jellyfish_path=self.input()["jellyfish"].path,
            alignment_path=self.input()["alignment"]["table"].path
            if self.is_blast_required
            else None,
        )
        sequences = self.prettify_sequences(df)
        config = self.prettify_configuration()

        util.log_and_check_candidates(self.logger, "CleanUpOutput", len(sequences))
        df.to_csv(self.output()["table"].path, index=False)
        Bio.SeqIO.write(sequences, self.output()["fasta"].path, format="fasta")
        with open(self.output()["config"].path, "w") as f:
            f.write(config)
        self.logger.info(f'Saving all files using hash "{util.get_gene_name()}"')

        # Files to be deleted
        if not RunConfig().save_intermediates:
            self.remove_intermediates()



"""Setup file for pypi package called efishent."""

# python setup.py sdist
# twine upload dist/latest-version.tar.gz

import textwrap
from setuptools import find_packages
from setuptools import setup

setup(
    # Description
    name="eFISHent",
    version="0.0.2",
    license="MIT",
    description="RNA FISH oligos/probes design tool.",
    long_description_content_type="text/plain",
    long_description=textwrap.dedent(
        """
    eFISHent is a command-line based tool to facilitate the creation of eFISHent RNA FISH oligonucleotide probes.
    This is just the python code which is reliant on multiple compiled bioinformatics libaries.
    Therefore please install other dependencies separately or use the bioconda package.
    """
    ),
    # Installation
    python_requires=">=3.7",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "biopython",
        "gtfparse",
        "luigi",
        "matplotlib",
        "numpy",
        "pandas",
        "pyarrow",
        "pyomo",
        "pysam",
    ],
    entry_points={
        "console_scripts": [
            "efishent = eFISHent.cli:main",
            "eFISHent = eFISHent.cli:main",
        ]
    },
    # Metadata
    author="Bastian Eichenberger",
    author_email="bastian@eichenbergers.ch",
    url="https://github.com/bbquercus/efishent/",
    project_urls={
        "Documentation": "https://github.com/BBQuercus/efishent/wiki",
        "Changelog": "https://github.com/BBQuercus/efishent/releases",
        "Issue Tracker": "https://github.com/bbquercus/efishent/issues",
    },
    keywords=["biomedical", "bioinformatics", "RNA probe design"],
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Utilities",
    ],
)



"""Luigi configurations passed as command line arguments."""

import luigi


class GeneralConfig(luigi.Config):
    reference_genome = luigi.Parameter(
        description="Path to reference genome fasta/fa file.",
        significant=True,
        default="",
    )
    reference_annotation = luigi.Parameter(
        description="Path to reference genome annotation (gene transfer format / gtf) file.",
        default="",
    )
    threads = luigi.IntParameter(description="Number of threads to launch.", default=2)
    output_dir = luigi.Parameter(
        description=(
            "Path to output directory. "
            "If not specified, will use the current directory."
        ),
        default="",
    )


class RunConfig(luigi.Config):
    build_indices = luigi.BoolParameter(
        description="Build indices for reference genome only.", default=False
    )
    analyze_probeset = luigi.Parameter(
        description=(
            "Path to the workflow's output fasta file. "
            "Will analyze the probe set in more detail and create a PDF. "
            "Note: reference genome, target gene of interest, "
            "and (optional) count table will still have to be provided!"
        ),
        default="",
    )
    save_intermediates = luigi.BoolParameter(
        description="Save intermediate files?", default=False
    )
    optimization_method = luigi.Parameter(
        description=(
            "Optimization method to use. "
            "Greedy will procedurally select the next longest probe. "
            "Optimal will optimize probes for maximum gene coverage - "
            "slow if many overlaps are present (typically with non-stringent parameter settings). "
            "[options: optimal, greedy]."
        ),
        default="greedy",
    )
    optimization_time_limit = luigi.IntParameter(
        description=(
            "Time limit for optimization in seconds. "
            "Only used if optimization method is set to 'optimal'."
        ),
        default=60,
    )


class SequenceConfig(luigi.Config):
    sequence_file = luigi.Parameter(
        description=(
            "Path to the gene's sequence file. "
            "Use this if the sequence can't be easily downloaded or if only certain regions should be targeted."
        ),
        default="",
    )
    ensembl_id = luigi.Parameter(
        description=(
            "Ensembl ID of the gene of interest. "
            "Can be used instead of gene and organism name to download the gene of interest. "
            "Used to filter out the gene of interest from FPKM based filtering - "
            "will do an automated blast-based filtering if not passed."
        ),
        default="",
    )
    gene_name = luigi.Parameter(description="Gene name.", default="")
    organism_name = luigi.Parameter(
        description=(
            "Latin name of the organism. "
            "Can be passed together with `ensembl_id` to narrow search."
        ),
        default="",
    )
    is_plus_strand = luigi.BoolParameter(
        description=(
            "Is the probe targeting the plus strand? "
            "Note: Entrez downloads will download the gene 5'-3'. "
            "Specifying the strand is therefore not required."
        ),
        default=True,
    )
    is_endogenous = luigi.BoolParameter(
        description="Is the probe endogenous?", default=True
    )


class ProbeConfig(luigi.Config):
    min_length = luigi.IntParameter(
        description="Minimum probe length in nucleotides.", default=21
    )
    max_length = luigi.IntParameter(
        description="Maximum probe length in nucleotides.", default=25
    )
    spacing = luigi.IntParameter(
        description="Minimum distance in nucleotides between probes.", default=2
    )
    min_tm = luigi.FloatParameter(
        description=(
            "Minimum melting temperature in degrees C. "
            "Formamide and Na concentration will affect melting temperature!"
        ),
        default=40.0,
    )
    max_tm = luigi.FloatParameter(
        description="Maximum melting temperature in degrees C (see minimum).",
        default=60.0,
    )
    min_gc = luigi.FloatParameter(
        description="Minimum GC content in percentage.", default=20.0
    )
    max_gc = luigi.FloatParameter(
        description="Maximum GC content in percentage.", default=80.0
    )
    formamide_concentration = luigi.FloatParameter(
        description="Formamide concentration as a percentage of the total buffer.",
        default=10.0,
    )
    na_concentration = luigi.FloatParameter(
        description="Na concentration in mM.", default=330.0
    )
    max_off_targets = luigi.IntParameter(
        description=(
            "Maximum number of off-targets binding anywhere BUT "
            "the gene of interest in the genome."
        ),
        default=0,
    )
    encode_count_table = luigi.Parameter(
        description=(
            "Path to the ENCODE RNAseq count table provided as TSV, CSV, or TXT format. "
            "The first two columns have to be `gene_id` and `normalized_value` respectively. "
            "Specific column names are not required. "
            "`gene_id` have to be ensemble IDs of the respective genes. "
            "`normalized_value` have to be normalized count values (RPKM, FPKM, TPM, DeSeq2-norm, etc). "
        ),
        default="",
    )
    max_expression_percentage = luigi.FloatParameter(
        description=(
            "Remove any off-target genes expressed more highly than "
            "`max-expression-percentage` percentage of all genes. "
            "Genes to remove are based on ENCODE RNAseq count table provided in `encode-count-table`. "
            "Only used if `encode-count-table` and `reference-annotation` are provided."
        ),
        default=50.0,
    )
    kmer_length = luigi.IntParameter(
        description="Length of k-mer used to filter probe sequences.", default=15
    )
    max_kmers = luigi.IntParameter(
        description="Highest count of sub-k-mers found in reference genome.", default=5
    )
    max_deltag = luigi.FloatParameter(
        description=(
            "Maximum predicted deltaG in kcal/mol. "
            "Note: deltaGs range from 0 (no secondary structures) to increasingly negative values!"
        ),
        default=-10.0,
    )
    sequence_similarity = luigi.IntParameter(
        description=(
            "Maximum percentage probes can be similar. "
            "Applied to their complements and reverse complements only. "
            "Ensures probes don't falsely bind to eachother. "
            "Setting any value above 0 could add multiple minute of runtime! "
            "The lower the value (above 0) the fewer potential probes."
        ),
        default=0,
    )



"""Filter probes based on alignment score and uniqueness."""

from typing import Any
import logging
import os
import subprocess

import Bio.SeqIO
import luigi
import numpy as np
import pandas as pd
import pysam

from . import constants
from . import util
from .basic_filtering import BasicFiltering
from .config import GeneralConfig
from .config import ProbeConfig
from .config import SequenceConfig
from .indexing import BuildBlastDatabase
from .indexing import BuildBowtieIndex
from .indexing import PrepareAnnotationFile
from .prepare_sequence import PrepareSequence


class AlignProbeCandidates(luigi.Task):
    """Align all probe candidates to the reference genome and filter."""

    logger = logging.getLogger("custom-logger")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_blast_required = (
            ProbeConfig().encode_count_table and SequenceConfig().is_endogenous
        )

    def requires(self):
        tasks = {"probes": BasicFiltering(), "bowtie": BuildBowtieIndex()}
        if self.is_blast_required:
            tasks["blastseq"] = PrepareSequence()
            tasks["gtf"] = PrepareAnnotationFile()
            tasks["blast"] = BuildBlastDatabase()
        return tasks

    def output(self):
        tasks = {
            "fasta": luigi.LocalTarget(
                os.path.join(
                    util.get_output_dir(), f"{util.get_gene_name()}_aligned.fasta"
                )
            )
        }
        if self.is_blast_required:
            tasks["table"] = luigi.LocalTarget(
                os.path.join(
                    util.get_output_dir(), f"{util.get_gene_name()}_counts.csv"
                )
            )
        return tasks

    @property
    def count_table(self) -> pd.DataFrame:
        """Read and verify a RNAseq normalized count table."""
        extension = os.path.splitext(self.fname_count)[1].lower()
        if extension not in (".tsv", ".csv", ".txt"):
            raise ValueError(
                "The count table must be provided as TSV, CSV, or TXT file. "
                f"Only found - {self.fname_count}"
            )
        sep = "\t" if extension in (".tsv", ".txt") else ","
        df = pd.read_csv(self.fname_count, sep=sep, index_col=0).reset_index().dropna()
        if not len(df.columns) >= 2:
            raise ValueError(
                "The count table must contain at least 2 columns."
                f"Only found - {df.columns}"
            )
        self.logger.debug(f"Mapping original the first two columns of - {df.columns}.")
        self.logger.debug(f"Read count table with {len(df)} entries.")
        self.logger.debug(f"Found dtypes - {df.dtypes}")
        self.logger.debug(f"First column - {df.iloc[0]}")

        df = df.iloc[:, :2]
        df.columns = ["gene_id", "count"]
        try:
            df = df.astype({"gene_id": str, "count": float})
        except ValueError:
            raise ValueError(
                "Could not convert count table columns to the right type. "
                "Please ensure the first two colums are of type string and number/float/int, respectively. "
            )

        # Remove potential version numbers
        df["gene_id"] = df["gene_id"].apply(lambda x: x.split(".")[0])
        return df

    @property
    def gtf_table(self) -> pd.DataFrame:
        """Parse prepared parquet-formatted gtf file."""
        return pd.read_parquet(self.fname_gtf)[constants.GTF_COLUMNS]

    @property
    def norm_table(self) -> pd.DataFrame:
        """Merge gtf and count table."""
        df = pd.merge(
            self.count_table, self.gtf_table, how="left", on="gene_id"
        ).dropna()
        return df

    def align_probes(self, max_off_targets: int, is_endogenous: bool, threads: int):
        """Align probes to the reference genome.

        Takes a fasta file as input (which will be converted to fastq).
        Saves a sam file with alignments.
        """
        # Convert fasta to fastq - bowtie doesn't return read names if not in fastq...
        fname_fastq = self.fname_fasta.rstrip("a") + "q"
        with open(self.fname_fasta, "r") as fasta, open(fname_fastq, "w") as fastq:
            for record in Bio.SeqIO.parse(fasta, "fasta"):
                record.letter_annotations["phred_quality"] = [40] * len(record)
                Bio.SeqIO.write(record, fastq, format="fastq")
        self.logger.debug(f"Converted fasta to fastq - {fname_fastq}")

        # Actual alignment with -m to only return alignments with >m hits
        alignments = str(max(max_off_targets + 1, 1))
        endo_exo_param = [] if is_endogenous else ["-m", alignments]
        args_bowtie = [
            "bowtie",
            self.fname_genome,
            fname_fastq,
            "--threads",
            str(threads),
            *endo_exo_param,
            "--sam",
            self.fname_sam,
        ]
        self.logger.debug(f"Running bowtie with - {' '.join(args_bowtie)}")
        subprocess.check_call(
            args_bowtie, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )

    @staticmethod
    def parse_raw_pysam(sam: str, is_endogenous: bool) -> pd.DataFrame:
        """Convert string based pysam output to a DataFrame."""
        # Parse tab and newline delimited pysam output
        data = [row.split("\t")[:10] for row in sam.split("\n")]
        # end = 14 if is_endogenous else 12
        columns = constants.SAMFILE_COLUMNS[:10]

        # Padding in case some samfiles are larger, first cols will stay the same
        # columns.extend(["" for _ in range(len(columns) - len(data[0]))])

        # If empty [['']]
        if data[0][0]:
            return pd.DataFrame(data, columns=columns).dropna()
        return pd.DataFrame(columns=columns)

    def filter_unique_probes(self, is_endogenous: bool) -> pd.DataFrame:
        """Filter sam file with probes based on alignment score and uniqueness."""
        # 60 - uniquely mapped read, regardless of number of mismatches / indels
        # 4 – flag for unmapped read
        flags = ["--min-MQ", "60"] if is_endogenous else ["--require-flags", "4"]
        self.logger.debug(f"Running samtools with - {' '.join(flags)}")
        filtered_sam = pysam.view(self.fname_sam, *flags)  # type: ignore
        return self.parse_raw_pysam(filtered_sam, is_endogenous)

    def exclude_gene_of_interest(
        self, df: pd.DataFrame, ensembl_id: str, fname_full_gene: str, threads: int
    ) -> pd.DataFrame:
        """Filter count table to exclude the gene of interest."""
        # Filter using provided EnsemblID directly
        if ensembl_id:
            self.logger.debug(f"Filtering directly using EmsembleID {ensembl_id}")
            return df[df["gene_id"] != ensembl_id]

        # Filter using blast
        args_blast = [
            "blastn",
            "-db",
            self.fname_genome,
            "-query",
            fname_full_gene,
            "-task",
            "megablast",
            "-outfmt",
            "6",
            "-num_threads",
            str(threads),
        ]
        self.logger.debug(f"Running blast with - {' '.join(args_blast)}")
        process = subprocess.run(args_blast, capture_output=True, text=True)
        data = [row.split("\t") for row in process.stdout.split("\n") if row]
        df_blast = pd.DataFrame(data, columns=constants.BLAST_COLUMNS).astype(
            {name: float for name in ["pident", "sstart", "send", "evalue"]}
        )

        # Arbirary identity values to ensure similar targets aren't falsely selected
        df_blast = df_blast[
            (df_blast["pident"] >= 98)
            & (df_blast["evalue"] <= 1e-5)
            & (df_blast["sseqid"] == df_blast.loc[0, "sseqid"])
        ]

        # Pad blast start/end results in case not everything was found/isoforms/etc.
        # Set to an arbitrary maximum value of 100 to not include other genes
        # Account for forward and reverse direction
        for buffer in range(0, 110, 10):
            gene_ids = []
            for _, row in df_blast.iterrows():
                gene_ids.extend(
                    df[
                        (
                            (df["seqname"] == row["sseqid"])
                            | (df["seqname"] == f"chr{row['sseqid']}")
                        )
                        & ((df["start"] >= row["sstart"]) & (df["end"] <= row["send"]))
                        | ((df["start"] >= row["send"]) & (df["end"] <= row["sstart"]))
                    ]["gene_id"].values
                )
            if not gene_ids:
                self.logger.debug(f"No genes found using a buffer of {buffer}")
                continue

            most_frequent_gene_id = max(set(gene_ids), key=gene_ids.count)
            self.logger.debug(
                f"Filtering out most frequent gene {most_frequent_gene_id}"
            )
            df = df[df["gene_id"] != most_frequent_gene_id].reset_index(drop=True)
            break

        return df

    def join_alignment_with_annotation(
        self, df_sam: pd.DataFrame, df_norm: pd.DataFrame
    ) -> pd.DataFrame:
        """Merge probe alignment with normalized count table based on start/end positions."""
        dfs = []

        for sequence, df_sequence in df_sam.groupby("rname"):
            df_count_sequence = df_norm[
                (df_norm["seqname"] == sequence)
                | (df_norm["seqname"] == sequence.lstrip("chr"))
            ]
            gene_start = df_count_sequence["start"].astype(int).values
            gene_end = df_count_sequence["end"].astype(int).values
            align_start = df_sequence["pos"].astype(int).values
            align_length = df_sequence["seq"].apply(len).values
            align_strand = (
                (df_sequence["flag"] == constants.SAM_FLAG_REVERSE)
                .replace({True: -1, False: 1})
                .values
            )
            align_end = align_start + align_strand * align_length

            # Create n*m matrix of which alignments(n) match up with which annotations(m)
            i, j = np.where(
                (
                    (align_start[:, None] >= gene_start)
                    & (align_start[:, None] <= gene_end)
                )
                | (
                    (align_end[:, None] >= gene_start)
                    & (align_end[:, None] <= gene_end)
                )
            )

            # Join columns to merge alignment information with count values
            df_sequence = pd.concat(
                [
                    df_sequence.reset_index(drop=True).iloc[i].reset_index(drop=True),
                    df_count_sequence.reset_index(drop=True)
                    .iloc[j]
                    .reset_index(drop=True),
                ],
                axis=1,
            )
            dfs.append(df_sequence)
        dfs = pd.concat(dfs, ignore_index=True)
        return dfs

    def get_most_expressed_genes(self, df: pd.DataFrame, percentage: float) -> Any:
        """Find the most highly expressed genes based on normalized count."""
        quantile = df["count"].quantile(percentage / 100)
        most_expressed = df[df["count"] > quantile].reset_index(drop=True)
        return most_expressed["gene_id"].values.tolist()

    # TODO allow filtering if exogenous?
    def filter_using_count(self, df_sam: pd.DataFrame) -> pd.DataFrame:
        """Filter candidates binding off targets above a expression percentage threshold."""
        df = self.join_alignment_with_annotation(df_sam=df_sam, df_norm=self.norm_table)
        df.to_csv(self.output()["table"].path, index=False)
        df_off_targets = self.exclude_gene_of_interest(
            df,
            ensembl_id=SequenceConfig().ensembl_id,
            fname_full_gene=self.input()["blastseq"].path,
            threads=GeneralConfig().threads,
        )
        most_expressed = self.get_most_expressed_genes(
            df=df_off_targets, percentage=ProbeConfig().max_expression_percentage
        )
        df_sam = df[~df["gene_id"].isin(most_expressed)].reset_index(drop=True)
        return df_sam

    def run(self):
        # Naming
        self.fname_fasta = self.input()["probes"].path
        self.fname_sam = os.path.splitext(self.fname_fasta)[0] + ".sam"
        self.fname_genome = util.get_genome_name()

        # Alignment and filtering
        self.align_probes(
            max_off_targets=ProbeConfig().max_off_targets,
            is_endogenous=SequenceConfig().is_endogenous,
            threads=GeneralConfig().threads,
        )
        df_sam = self.filter_unique_probes(is_endogenous=SequenceConfig().is_endogenous)
        if self.is_blast_required:
            self.fname_gtf = self.input()["gtf"].path
            self.fname_count = ProbeConfig().encode_count_table
            df_sam = self.filter_using_count(df_sam)

        # Save output
        sequences = list(Bio.SeqIO.parse(self.fname_fasta, "fasta"))
        candidates = [seq for seq in sequences if seq.id in df_sam["qname"].unique()]
        self.logger.debug(df_sam["qname"].values)
        util.log_and_check_candidates(
            self.logger, "AlignProbeCandidates", len(candidates), len(sequences)
        )
        Bio.SeqIO.write(candidates, self.output()["fasta"].path, format="fasta")
